"""Utility functions for experiment definitions."""

import beobench.utils


def get_experiment_config(
    problem_def: dict, method_def: dict, rllib_setup_def: dict
) -> dict:
    """Combine partial ray tune experiment definitions into single definition.

    This recursively merges the dictionaries.

    Additional info: note that RLlib is a submodule of the ray package, i.e. it is
    imported as `ray.rllib`. For experiment definitions it uses the `ray.tune`
    submodule. Therefore ray tune experiment definition means the same as ray rllib
    experiment defintions. To avoid confusion all variable/argument names use rllib
    instead of ray tune but strictly speaking these are ray tune experiment
    definitions.

    Args:
        problem_def (dict): definition of problem. This is an incomplete
            ray tune experiment defintion that only defines the problem side.
        method_def (dict): definition of method. This is an incomplete
            ray tune experiment defintion that only defines the method side.
        rllib_setup (dict): rllib setup. This is an incomplete
            ray tune experiment defintion that only defines the ray tune/rllib setup
            (e.g. number of workers, etc.).

    Returns:
        dict: combined dictionary.
    """

    exp_config = {}

    exp_config = beobench.utils.merge_dicts(
        problem_def["rllib_experiment_config"], method_def["rllib_experiment_config"]
    )
    exp_config = beobench.utils.merge_dicts(
        exp_config, rllib_setup_def["rllib_experiment_config"]
    )

    exp_config["name"] = problem_def["name"] + "_" + method_def["name"]

    # Get combine the two dict into single experiment
    # conf file for ray.tune.run()
    return exp_config



"""Default experiment definitions."""

import ray.tune

import beobench.integrations.boptest
import beobench.utils

PROBLEM_001_BOPTEST_HEATPUMP = {
    "name": "problem_001",
    "description": (
        "Control problem corresponding to "
        "the BOPTEST testcase 'bestest_hydronic_heat_pump'."
    ),
    "env_creator_to_register": beobench.integrations.boptest.create_env,
    "rllib_experiment_config": {
        "config": {
            "env": "beobench.integrations.boptest.create_env",
            "env_config": {
                "boptest_testcase": "bestest_hydronic_heat_pump",
                "gym_kwargs": {
                    "actions": ["oveHeaPumY_u"],
                    "observations": {"reaTZon_y": (280.0, 310.0)},
                    "random_start_time": True,
                    "max_episode_length": 86400,
                    "warmup_period": 10,
                    "step_period": 900,
                },
                "normalize": True,
            },
            "gamma": 0.999,
        },
        "stop": {"training_iteration": 50},
    },
}

METHOD_001_PPO = {
    "name": "PPO",
    "description": "Proximal policy optimisation.",
    "rllib_experiment_config": {
        "run_or_experiment": "PPO",
        "config": {
            "lr": 5e-3,
            "model": {
                "fcnet_hiddens": [256, 256, 256, 256],
                "fcnet_activation": "relu",
                "post_fcnet_activation": "tanh",
            },
        },
    },
}

RLLIB_SETUP = {
    "rllib_experiment_config": {
        "config": {
            # Utilities settings
            "framework": "torch",
            "log_level": "WARNING",
            "num_workers": ray.tune.grid_search(
                [1, 4, 6, 10, 12]
            ),  # 1 for silent mode, can at least be 6
            "num_gpus": 1,
            "seed": ray.tune.randint(0, 10000000),
        },
        "local_dir": "./tmp/tune/",
        "log_to_file": True,
        "checkpoint_freq": 10000,
        "num_samples": 1,
    },
}


problem = PROBLEM_001_BOPTEST_HEATPUMP
method = METHOD_001_PPO
rllib_setup = RLLIB_SETUP



"""Module with a number of utility functions."""

import docker


def check_if_in_notebook() -> bool:
    """Check if code is executed from jupyter notebook.

    Taken from https://stackoverflow.com/a/44100805.

    Returns:
        bool: whether code is run from notebook.
    """
    try:
        # This function should only be available in ipython kernel.
        get_ipython()  # pylint: disable=undefined-variable
        return True
    except:  # pylint: disable=bare-except
        return False


def merge_dicts(a: dict, b: dict, path: list = None, mutate_a: bool = False) -> dict:
    """Merge dictionary b into dictionary a.

    Adapted from https://stackoverflow.com/a/7205107.

    Args:
        a (dict): a dicitonary
        b (dict): another dictionary
        path (list, optional): where the dict is in the original dict.
            Necessary for recursion, no need to use. Defaults to None.
        mutate_a (bool, optional): whether to mutate the dictionary a that
            is given. Necessary for recursion, no need to use.
            Defaults to False.


    Raises:
        Exception: When dictionaries are inconsistent

    Returns:
        dictionary: merged dictionary.
    """
    # pylint: disable=consider-using-f-string

    # Ensure that dict a is not mutated
    if not mutate_a:
        a = dict(a)

    if path is None:
        path = []
    for key in b:
        if key in a:
            if isinstance(a[key], dict) and isinstance(b[key], dict):
                merge_dicts(a[key], b[key], path + [str(key)], mutate_a=True)
            elif a[key] == b[key]:
                pass  # same leaf value
            else:
                raise Exception("Conflict at %s" % ".".join(path + [str(key)]))
        else:
            a[key] = b[key]
    return a


def shutdown() -> None:
    """Shut down all beobench and BOPTEST containers."""

    print("Stopping any remaining beobench and BOPTEST docker containers...")

    client = docker.from_env()
    container_num = 0
    for container in client.containers.list():
        if "auto_beobench" in container.name or "auto_boptest" in container.name:
            print(f"Stopping container {container.name}")
            container.stop(timeout=0)
            container_num += 1

    print(f"Stopped {container_num} container(s).")


def restart() -> None:
    """Clean up remaining beobench processes and containers
    before running new experiments.

    This stops all docker containers still running. This
    function is not called by other scheduler functions
    to enable the parallel running of experiments.
    """

    shutdown()



"""An experiment to test sinergym integration."""

import ray.tune
from beobench.experiment.definitions.default import problem, method, rllib_setup


problem = {
    "name": "sinergym_test_problem",
    "description": ("Control problem corresponding to " "created by sinergym env ''."),
    "problem_library": "sinergym",
    "rllib_experiment_config": {
        "config": {
            "env": "sinergym-Eplus-5Zone-hot-continuous-v1",
            "env_config": {
                "name": "Eplus-5Zone-hot-continuous-v1",
                "normalize": True,
            },
            "gamma": 0.999,
            "output": "/root/ray_results/debug/",
            "output_compress_columns": [],
            "soft_horizon": False,
            "no_done_at_end": False,
            "horizon": 10000,
        },
        "stop": {"timesteps_total": 400000},
    },
}

rllib_setup = {
    "rllib_experiment_config": {
        "config": {
            # Utilities settings
            "framework": "torch",
            "log_level": "WARNING",
            "num_workers": 8,  # 1 for silent mode, can at least be 6
            "num_gpus": 1,
            "seed": ray.tune.randint(0, 10000000),
        },
        "log_to_file": True,
        "checkpoint_freq": 10000,
    },
}



"""Module to schedule experiments."""

import ray.tune
import ray.tune.integration.wandb
import click
import uuid
import subprocess
import pathlib
import importlib.util
import os

import beobench.experiment.definitions.utils
import beobench.experiment.definitions.default
import beobench.experiment.containers
import beobench.utils


def run(
    experiment_file: str = None,
    local_dir: str = "./beobench_results/ray_results",
    wandb_project: str = "",
    wandb_entity: str = "",
    wandb_api_key: str = "",
    use_gpu: bool = False,
    docker_shm_size: str = "2gb",
    no_additional_container: bool = False,
    use_no_cache: bool = False,
) -> None:
    """Run experiment.

    This function allows the use to run experiments from the command line or python
    interface.

    Args:
        experiment_file (str, optional): File that defines experiment.
            Defaults to None.
        local_dir (str, optional): Directory to write experiment files to. This argument
            is equivalent to the `local_dir` argument in `tune.run()`. Defaults to
            `"./beobench_results/ray_results"`.
        wandb_project (str, optional): Name of wandb project. Defaults to
            "initial_experiments".
        wandb_entity (str, optional): Name of wandb entity. Defaults to "beobench".
        wandb_api_key (str, optional): wandb API key. Defaults to None.
        use_gpu (bool, optional): whether to use GPU from the host system. Defaults to
            False.
        docker_shm_size(str, optional): size of the shared memory available to the
            container. Defaults to '2gb'."
        no_additional_container (bool, optional): wether not to start another container
            to run experiments in. Defaults to False, which means that another container
            is started to run experiments in.
        use_no_cache (bool, optional): whether to use cache to build experiment
            container.
    """

    # Load experiment definition file
    if experiment_file is None:
        experiment_def = beobench.experiment.definitions.default
    else:
        experiment_file = pathlib.Path(experiment_file)
        # import experiment definition file as module
        spec = importlib.util.spec_from_file_location(
            "experiment_definition",
            str(experiment_file.absolute()),
        )
        experiment_def = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(experiment_def)

        # check if all attributes present and replace with default if not
        if not hasattr(experiment_def, "problem"):
            experiment_def.problem = beobench.experiment.definitions.default.problem
        if not hasattr(experiment_def, "method"):
            experiment_def.method = beobench.experiment.definitions.default.method
        if not hasattr(experiment_def, "rllib_setup"):
            experiment_def.rllib_setup = (
                beobench.experiment.definitions.default.rllib_setup
            )

    if no_additional_container:

        # Add wandb callback if sufficient information
        # TODO: add earlier check to see that also API key available
        if wandb_project and wandb_entity:
            callbacks = [_create_wandb_callback(wandb_project, wandb_entity)]
        elif wandb_project or wandb_entity:
            raise ValueError(
                "Only one of wandb_project or wandb_entity given, but both required."
            )
        else:
            callbacks = []

        # change RLlib setup if GPU used
        rllib_setup = experiment_def.rllib_setup
        if use_gpu:
            rllib_setup["rllib_experiment_config"]["config"]["num_gpus"] = 1

        # run experiment in ray tune
        run_in_tune(
            problem_def=experiment_def.problem,
            method_def=experiment_def.method,
            rllib_setup=rllib_setup,
            rllib_callbacks=callbacks,
        )
    else:
        # build and run experiments in docker container

        # Ensure local_dir exists, and create otherwise
        local_dir_path = pathlib.Path(local_dir)
        local_dir_path.mkdir(parents=True, exist_ok=True)
        local_dir_abs = str(local_dir_path.absolute())

        # docker setup
        image_tag = beobench.experiment.containers.build_experiment_container(
            build_context=experiment_def.problem["problem_library"],
            use_no_cache=use_no_cache,
        )

        # define docker arguments/options/flags
        unique_id = uuid.uuid4().hex[:6]
        container_name = f"auto_beobench_experiment_{unique_id}"

        docker_flags = []
        if experiment_file is not None:
            exp_file_abs = experiment_file.absolute()
            exp_file_on_docker = f"/tmp/beobench/{experiment_file.name}"
            docker_flags += [
                "-v",
                f"{exp_file_abs}:{exp_file_on_docker}:ro",
            ]

        # enable docker-from-docker access only for built-in boptest integration.
        if experiment_def.problem["problem_library"] == "boptest":

            # Create docker network (only useful if starting other containers)
            beobench.experiment.containers.create_docker_network("beobench-net")

            docker_flags += [
                # enable access to docker-from-docker
                "-v",
                "/var/run/docker.sock:/var/run/docker.sock",
                # network allows access to BOPTEST API in other containers
                "--network",
                "beobench-net",
            ]

        # enabling GPU access in docker container
        if use_gpu:
            docker_flags += [
                # add all available GPUs
                "--gpus=all",
            ]

        # define flags for beobench scheduler call inside experiment container
        beobench_flags = []
        if experiment_file:
            beobench_flags.append(f"--experiment-file={exp_file_on_docker}")
        if wandb_project:
            beobench_flags.append(f"--wandb-project={wandb_project}")
        if wandb_entity:
            beobench_flags.append(f"--wandb-entity={wandb_entity}")
        if use_gpu:
            beobench_flags.append("--use-gpu")
        beobench_flag_str = " ".join(beobench_flags)

        # if no wandb API key is given try to get it from env
        if not wandb_api_key:
            # this will return "" if env var not set
            wandb_api_key = os.getenv("WANDB_API_KEY", "")

        args = [
            "docker",
            "run",
            # mount experiment data dir
            "-v",
            f"{local_dir_abs}:/root/ray_results",
            # automatically remove container when stopped/exited
            "--rm",
            # add more memory
            f"--shm-size={docker_shm_size}",
            "--name",
            container_name,
            *docker_flags,
            image_tag,
            "/bin/bash",
            "-c",
            (
                f"export WANDB_API_KEY={wandb_api_key} && python -m "
                f"beobench.experiment.scheduler {beobench_flag_str} "
                "--no-additional-container && bash"
            ),
        ]
        print("Executing docker command: ", " ".join(args))
        subprocess.check_call(args)


@click.command()
@click.option(
    "--experiment-file",
    default=None,
    help="File that defines beobench experiment.",
    type=click.Path(exists=True, file_okay=True, dir_okay=False),
)
@click.option(
    "--local-dir",
    default="./beobench_results/ray_results",
    help="Local directory to write results to.",
    type=click.Path(exists=False, file_okay=False, dir_okay=True),
)
@click.option(
    "--wandb-project",
    default="",
    help="Weights and biases project name to log runs to.",
)
@click.option(
    "--wandb-entity",
    default="",
    help="Weights and biases entity name to log runs under.",
)
@click.option(
    "--wandb-api-key",
    default="",
    help="Weights and biases API key.",
)
@click.option(
    "--use-gpu",
    is_flag=True,
    help="Whether to use GPUs from the host system in experiment container.",
)
@click.option(
    "--docker-shm-size",
    default="2gb",
    help="Size of shared memory available to experiment container.",
)
@click.option(
    "--no-additional-container",
    is_flag=True,
    help="Do not run another container to do experiments in.",
)
@click.option(
    "--use-no-cache",
    is_flag=True,
    help="Whether to use cache to build experiment container.",
)
def run_command(
    experiment_file: str,
    local_dir: str,
    wandb_project: str,
    wandb_entity: str,
    wandb_api_key: str,
    use_gpu: bool,
    docker_shm_size: str,
    no_additional_container: bool,
    use_no_cache: bool,
) -> None:
    """Run experiment from command line.

    Command line version of run function. This appears to be
    the best (but not great) way to have a parallel python and command
    line interface.

    See https://stackoverflow.com/a/40094408.
    """
    run(
        experiment_file=experiment_file,
        local_dir=local_dir,
        wandb_project=wandb_project,
        wandb_entity=wandb_entity,
        wandb_api_key=wandb_api_key,
        use_gpu=use_gpu,
        docker_shm_size=docker_shm_size,
        no_additional_container=no_additional_container,
        use_no_cache=use_no_cache,
    )


def run_in_tune(
    problem_def: dict, method_def: dict, rllib_setup: dict, rllib_callbacks: list = None
) -> ray.tune.ExperimentAnalysis:
    """Run beobench experiment.

    Additional info: note that RLlib is a submodule of the ray package, i.e. it is
    imported as `ray.rllib`. For experiment definitions it uses the `ray.tune`
    submodule. Therefore ray tune experiment definition means the same as ray rllib
    experiment defintions. To avoid confusion all variable/argument names use rllib
    instead of ray tune but strictly speaking these are ray tune experiment
    definitions.

    Args:
        problem_def (dict): definition of problem. This is an incomplete
            ray tune experiment defintion that only defines the problem side.
        method_def (dict): definition of method. This is an incomplete
            ray tune experiment defintion that only defines the method side.
        rllib_setup (dict): rllib setup. This is an incomplete
            ray tune experiment defintion that only defines the ray tune/rllib setup
            (e.g. number of workers, etc.).
        rllib_callbacks (list, optional): callbacks to add to ray.tune.run command.
            Defaults to None.

    Returns:
        ray.tune.ExperimentAnalysis: analysis object of completed experiment.
    """
    if rllib_callbacks is None:
        rllib_callbacks = []

    # combine the three incomplete ray tune experiment
    # definitions into a single complete one.
    exp_config = beobench.experiment.definitions.utils.get_experiment_config(
        problem_def, method_def, rllib_setup
    )

    # register the problem environment with ray tune
    # env_creator is a module available in experiment containers
    import env_creator  # pylint: disable=import-outside-toplevel

    ray.tune.registry.register_env(
        problem_def["rllib_experiment_config"]["config"]["env"],
        env_creator.create_env,
    )

    # if run in notebook, change the output reported throughout experiment.
    if beobench.utils.check_if_in_notebook():
        reporter = ray.tune.JupyterNotebookReporter(overwrite=True)
    else:
        reporter = None

    # running the experiment
    analysis = ray.tune.run(
        progress_reporter=reporter,
        callbacks=rllib_callbacks,
        **exp_config,
    )

    return analysis


def _create_wandb_callback(
    wandb_project: str,
    wandb_entity: str,
):
    wandb_callback = ray.tune.integration.wandb.WandbLoggerCallback(
        project=wandb_project, log_config=True, entity=wandb_entity
    )
    return wandb_callback


if __name__ == "__main__":
    run_command()  # pylint: disable=no-value-for-parameter



"""Module for managing experiment containers."""

import subprocess
import os


def build_experiment_container(
    build_context: str, docker_tag: str = None, use_no_cache: bool = False
) -> None:
    """Build experiment container from beobench/integrations/boptest/Dockerfile.

    Args:
        build_context (str): context to build from. This can either be a path to
            directory with Dockerfile in it, or a URL to a github repo, or name
            of existing beobench integration (e.g. `boptest`). See the official docs
            https://docs.docker.com/engine/reference/commandline/build/ for more info.
        use_no_cache (bool, optional): wether to use cache in build. Defaults to False.
    """

    flags = []
    if use_no_cache:
        flags.append("--no-cache")

    AVAILABLE_INTEGRATIONS = ["boptest","sinergym"]  # pylint: disable=invalid-name

    if build_context in AVAILABLE_INTEGRATIONS:
        docker_tag = f"beobench_{build_context}:latest"
        integration_name = build_context
        build_context = (
            f"https://github.com/rdnfn/"
            f"beobench_contrib.git#main:gyms/{build_context}"
        )
        print(
            (
                f"Recognised integration named {integration_name}: using build"
                f" context {build_context}"
            )
        )
    else:
        if docker_tag is None:
            # get alphanumeric name from context
            context_name = "".join(e for e in build_context if e.isalnum())
            docker_tag = f"beobench_custom_{context_name}:latest"

    print(f"Building experiment container `{docker_tag}`...")

    args = [
        "docker",
        "build",
        "-t",
        docker_tag,
        "-f",
        "Dockerfile",  # change to non-default name
        *flags,
        build_context,
    ]
    env = os.environ.copy()
    env["DOCKER_BUILDKIT"] = "0"
    subprocess.check_call(
        args,
        env=env,  # this enables accessing dockerfile in subdir
    )

    print("Experiment container build finished.")

    return docker_tag


def create_docker_network(network_name: str) -> None:
    """Create docker network.

    For more details see
    https://docs.docker.com/engine/reference/run/#network-settings

    Args:
        network_name (str): name of docker network.
    """

    print("Creating docker network ...")
    try:
        args = ["docker", "network", "create", network_name]
        subprocess.check_call(args)
        print("Docker network created.")
    except subprocess.CalledProcessError:
        print("No new network created. Network may already exist.")



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages

with open("PYPI_README.rst", encoding="UTF-8") as readme_file:
    readme = readme_file.read()

with open("HISTORY.rst", encoding="UTF-8") as history_file:
    history = history_file.read()

version = "0.2.1"

requirements = [
    "wandb",
    "docker",
    "ray[rllib]",
    "click",
    "torch",
    "gym",
]


test_requirements = []

setup(
    author="rdnfn",
    python_requires=">=3.6",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    description="Beobench is a toolbox for benchmarking reinforcement learning (RL) algorithms on building energy optimisation (BEO) problems.",  # pylint: disable=line-too-long
    install_requires=requirements,
    license="MIT license",
    long_description=readme + "\n\n" + history,
    include_package_data=True,
    keywords="beobench",
    name="beobench",
    packages=find_packages(include=["beobench", "beobench.*"]),
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/rdnfn/beobench",
    zip_safe=False,
    version=version,
)



"""Module with various configuration constants"""

import pathlib

# import os

# Path locations
# DEFAULT_INSTALL_PATH = pathlib.Path(os.getcwd() + "/tmp/beobench_external_install/")
DEFAULT_INSTALL_PATH = pathlib.Path("/opt/beobench")


# REPO URLS
BOPTEST_REPO_URL = "https://github.com/ibpsa/project1-boptest.git"
BOPTEST_REPO_NAME = "project1-boptest"
BOPTEST_COMMIT = "703daf26a9e9e18de7141c021cbfe36aa3578ea8"
BOPTEST_PIP_DEP = [
    "flask-restful==0.3.9",
    "pandas==1.3.4",
    "flask_cors==3.0.10",
    # BOPTEST gym cannot be added as a direct or extra
    # dependency in setup.py because pypi does not accept
    # such direct dependencies. Therefore it is installed
    # here.
    (
        "boptest_gym@git+"
        "https://github.com/rdnfn/project1-boptest-gym.git@rdnfn/feature-packaging"
    ),
]



"""Default experiment definitions."""

import ray.tune

import beobench.integrations.boptest
import beobench.utils
import beobench.experiment.definitions.default

problem = {
    "name": "problem_001",
    "description": (
        "Control problem corresponding to "
        "the BOPTEST testcase 'bestest_hydronic_heat_pump'."
    ),
    "env_creator_to_register": beobench.integrations.boptest.create_env,
    "rllib_experiment_config": {
        "config": {
            "env": "beobench.integrations.boptest.create_env",
            "env_config": {
                "boptest_testcase": "bestest_hydronic_heat_pump",
                "gym_kwargs": {
                    "actions": ["oveHeaPumY_u"],
                    "observations": {"reaTZon_y": (280.0, 310.0)},
                    "random_start_time": True,
                    "max_episode_length": 86400,
                    "warmup_period": 10,
                    "step_period": 900,
                },
                "discretize": ray.tune.choice([4, 10, 20, 40]),
            },
            "gamma": 0.999,
        },
        "stop": {"timesteps_total": 400000},
    },
}

method = {
    "name": "DQN",
    "description": "Deep Q-network.",
    "rllib_experiment_config": {
        "run_or_experiment": "DQN",
        "config": {
            "lr": ray.tune.choice([0.001, 0.0001, 0.00001]),
            "model": {
                "fcnet_hiddens": [256, 256, 256, 256],
                "fcnet_activation": ray.tune.choice(["tanh", "linear", "relu"]),
                "post_fcnet_activation": ray.tune.choice(["tanh", "linear"]),
            },
        },
        "num_samples": 20,
    },
}

rllib_setup = beobench.experiment.definitions.default.rllib_setup
